@@ -29,10 +29,7 @@ Additional planned extensions include:
     * make stack_size a kwarg when creating a thread
     * native events, semaphores and timed waits on win32
     * native conditions and timed waits on pthreads platforms
-
-Stuff that might get included one day:
-
-    * ReadWriteLock (using SRW on Win Vista+, pthread_rwlock on posix)
+    * native SHLock implementations (SRW on Win Vista+, pthread_rwlock)
 
 """
 
@@ -41,7 +38,7 @@ from __future__ import with_statement
 
 __ver_major__ = 0
 __ver_minor__ = 2
-__ver_patch__ = 0
+__ver_patch__ = 1
 __ver_sub__ = ""
 __version__ = "%d.%d.%d%s" % (__ver_major__,__ver_minor__,
                               __ver_patch__,__ver_sub__)
